"""Common exception module."""


class ArgumentError(ValueError):
    """Argument exception."""


class OneSampleError(Exception):
    """One sample exception."""


class UpdateDetectorError(Exception):
    """Update detector exception."""
